import dataclasses
from abc import abstractmethod
import typing
from typing import (
    Any,
    Generic,
    Iterable,
    Optional,
    Sequence,
    TypeVar,
)

from aiohttp import ClientSession
from pypika import (  # type: ignore
    ClickHouseQuery,
    Criterion,
    CustomFunction,
    Not,
    Order,
    Parameter,
    Table,
    functions as fn,
)
from pypika.terms import PseudoColumn  # type: ignore
from pypika.clickhouse.array import Array, HasAny, Field  # type: ignore
from pypika.clickhouse.search_string import Match  # type: ignore

from .core import (
    F,
    FilterType,
    NotFound,
    QueryError,
)


T = TypeVar("T")

_is_ip_address_in_range = CustomFunction("isIPAddressInRange", ["address", "prefix"])
_is_null = CustomFunction("isNull", ["x"])
_is_not_null = CustomFunction("isNotNull", ["x"])


class ClickHouseRepo(Generic[T]):
    """Implementation of the :class:`misery.Repo` protocol
    that uses aiohttp to communicate with ClickHouse.
    """

    id_field = "id"
    """Name of the ID field of an entity."""

    @property
    @abstractmethod
    def table(self) -> Table:
        """Main table of the repository.
        It is used to autogenerate SQL queries.
        """

    def __init__(self, session: ClientSession) -> None:
        """:param session: Client session for
        making HTTP requests to ClickHouse.
        """
        self.session = session

    @property
    def query(self) -> ClickHouseQuery:
        """Query to select records from the database.

        Override this property when you need
        something more complex than selecting
        all columns from the main table
        of the repository.
        """
        return ClickHouseQuery.from_(self.table).select("*")

    @property
    def _entity_type(self) -> type:
        return typing.get_args(self.__orig_bases__[0])[0]  # type: ignore

    def load(self, record: dict) -> T:
        """Map a database record to an entity.

        Override this method if needed.
        """
        return self._entity_type(**record)

    def dump(self, entity: T) -> dict:
        """Map an entity to a database record.

        Override this method if needed.
        """
        return dataclasses.asdict(entity)

    async def request_json(self, query: ClickHouseQuery) -> dict:
        query_str = str(query) + " FORMAT JSON"

        async with self.session.post(
            "/",
            data=query_str,
        ) as resp:
            if resp.status != 200:
                raise QueryError(
                    message=(await resp.read()).decode(),
                    query=query_str,
                )

            data = await resp.json()
            return data

    async def request(self, query: ClickHouseQuery) -> bytes:
        query_str = str(query)

        async with self.session.post(
            "/",
            data=query_str,
        ) as resp:
            if resp.status != 200:
                raise QueryError(
                    message=(await resp.read()).decode(),
                    query=query_str,
                )
            data = await resp.read()
            return data

    async def fetch_one(self, query: ClickHouseQuery) -> T:
        """Find a record in the database
        and map it to an entity.
        """
        data = await self.request_json(query)

        if data["rows"] < 1:
            raise NotFound

        return self.load(data["data"][0])

    async def fetch_many(self, query: ClickHouseQuery) -> Iterable[T]:
        """Find multiple records in the database
        and map them to entities.
        """
        data = await self.request_json(query)
        return map(self.load, data["data"])

    async def add(self, entity: T) -> None:
        data = self.dump(entity)
        query = (
            ClickHouseQuery.into(self.table)
            .columns(*data.keys())
            .insert(*data.values())
        )

        await self.request(query)

    async def add_many(self, entities: Iterable[T]) -> None:
        entities = list(entities)

        if len(entities) == 0:
            return

        data = self.dump(entities[0])

        query = (
            ClickHouseQuery.into(self.table)
            .columns(*data.keys())
            .insert(*[tuple(self.dump(e).values()) for e in entities])
        )

        await self.request(query)

    async def get(self, **kwargs: Any) -> T:
        query = self.query.limit(1)

        for k, v in kwargs.items():
            query = query.where(self.table[k] == v)

        return await self.fetch_one(query)

    async def get_for_update(self, **kwargs: Any) -> T:
        return await self.get(**kwargs)

    async def get_many(
        self,
        filters: Sequence[F] = (),
        order: Sequence[str] = (),
        limit: Optional[int] = None,
        page: int = 1,
        offset: int = 0,
    ) -> Iterable[T]:
        query = self._params_to_query(
            filters=filters,
            order=order,
            limit=limit,
            page=page,
            offset=offset,
        )

        return await self.fetch_many(query)

    def _params_to_query(
        self,
        filters: Sequence[F] = (),
        order: Sequence[str] = (),
        limit: Optional[int] = None,
        page: int = 1,
        offset: int = 0,
    ) -> ClickHouseQuery:
        query = self.query

        if limit is not None:
            query = query.limit(limit).offset((page - 1) * limit + offset)
        elif offset > 0:
            query = query.offset(offset)

        for field in order:
            if field.startswith("-"):
                query = query.orderby(field[1:], order=Order.desc)
            else:
                query = query.orderby(field)

        for f in filters:
            criterion = self._filter_to_criterion(f)
            query = query.where(criterion)

        return query

    def _filter_to_criterion(self, f: F) -> Criterion:
        if f.type == FilterType.OR:
            return Criterion.any([self._filter_to_criterion(ff) for ff in f.value])
        elif f.type == FilterType.AND:
            return Criterion.all([self._filter_to_criterion(ff) for ff in f.value])

        column = self.table[f.field]
        criterion: Any = None

        if f.type == FilterType.EQ:
            if f.value is None:
                criterion = _is_null(column)
            else:
                criterion = column == f.value
        elif f.type == FilterType.NEQ:
            if f.value is None:
                criterion = _is_not_null(column)
            else:
                criterion = column != f.value
        elif f.type == FilterType.LT:
            criterion = column < f.value
        elif f.type == FilterType.GT:
            criterion = column > f.value
        elif f.type == FilterType.LTE:
            criterion = column <= f.value
        elif f.type == FilterType.GTE:
            criterion = column >= f.value
        elif f.type == FilterType.STARTSWITH:
            criterion = column.like(f"{f.value}%")
        elif f.type == FilterType.NSTARTSWITH:
            criterion = column.not_like(f"{f.value}%")
        elif f.type == FilterType.ENDSWITH:
            criterion = column.like(f"%{f.value}")
        elif f.type == FilterType.NENDSWITH:
            criterion = column.not_like(f"%{f.value}")
        elif f.type == FilterType.ISTARTSWITH:
            criterion = column.ilike(f"{f.value}%")
        elif f.type == FilterType.NISTARTSWITH:
            criterion = column.not_ilike(f"{f.value}%")
        elif f.type == FilterType.IENDSWITH:
            criterion = column.ilike(f"%{f.value}")
        elif f.type == FilterType.NIENDSWITH:
            criterion = column.not_ilike(f"%{f.value}")
        elif f.type == FilterType.CONTAINS:
            criterion = column.like(f"%{f.value}%")
        elif f.type == FilterType.NCONTAINS:
            criterion = column.not_like(f"%{f.value}%")
        elif f.type == FilterType.ICONTAINS:
            criterion = column.ilike(f"%{f.value}%")
        elif f.type == FilterType.NICONTAINS:
            criterion = column.not_ilike(f"%{f.value}%")
        elif f.type == FilterType.IN:
            criterion = column.isin(f.value)
        elif f.type == FilterType.NIN:
            criterion = column.notin(f.value)
        elif f.type == FilterType.MATCHES:
            criterion = Match(column, f.value)
        elif f.type == FilterType.NMATCHES:
            criterion = Not(Match(column, f.value))
        elif f.type == FilterType.IMATCHES:
            criterion = Match(column, "(?i)" + f.value)
        elif f.type == FilterType.NIMATCHES:
            criterion = Not(Match(column, "(?i)" + f.value))
        elif f.type == FilterType.IPIN:
            criterion = _is_ip_address_in_range(column, f.value)
        elif f.type == FilterType.NIPIN:
            criterion = Not(_is_ip_address_in_range(column, f.value))
        elif f.type == FilterType.HASANY:
            criterion = HasAny(column, Array(list(f.value)))

        if f.not_:
            return Not(criterion)

        return criterion

    async def get_first(
        self, filters: Sequence[F] = (), order: Sequence[str] = ()
    ) -> T:
        try:
            return list(await self.get_many(filters, order=order, limit=1))[0]
        except IndexError:
            raise NotFound

    async def update(self, entity: T) -> None:
        """Don't use this method in production
        because ClickHouse isn't good at updating records.
        """

        old_entity = await self.get(**{self.id_field: getattr(entity, self.id_field)})
        old_record = self.dump(old_entity)
        record = self.dump(entity)
        query = ClickHouseQuery.update(self.table)

        for k, v in record.items():
            if old_record[k] != record[k]:
                query = query.set(k, v)

        query = query.where(self.table[self.id_field] == getattr(entity, self.id_field))
        await self.request(query)

    async def delete(self, **kwargs: Any) -> None:
        """Don't use this method in production
        because ClickHouse isn't good at deleting records.
        """

        query = ClickHouseQuery.from_(self.table).delete()

        if kwargs:
            for k, v in kwargs.items():
                query = query.where(self.table[k] == v)
        else:
            query = query.where(PseudoColumn(1) == 1)

        await self.request(query)

    async def exists(self, **kwargs: Any) -> bool:
        query = ClickHouseQuery.from_(self.table).select(1).limit(1)

        for k, v in kwargs.items():
            query = query.where(self.table[k] == v)

        data = await self.request_json(query)
        return data["rows"] > 0

    async def count(self, **kwargs: Any) -> int:
        query = ClickHouseQuery.from_(self.table).select(fn.Count("*"))

        for k, v in kwargs.items():
            query = query.where(self.table[k] == v)

        data = await self.request_json(query)
        return int(data["data"][0]["count()"])

    async def count_filtered(self, filter_: F) -> int:
        criterion = self._filter_to_criterion(filter_)
        query = self.query.where(criterion).distinct()
        query = ClickHouseQuery.from_(query).select(fn.Count(1))
        data = await self.request_json(query)
        return int(data["data"][0]["count()"])


class ClickHouseTransactionManager:
    """This class exists only for compatibility with the protocol
    because ClickHouse doesn't support transactions.
    It doesn't do anything to the database."""

    async def __aenter__(self) -> None:
        pass

    async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        pass



import dataclasses
import itertools
import typing
from abc import abstractmethod
from contextvars import ContextVar
from enum import Enum
from typing import (
    Any,
    Generic,
    Iterable,
    Optional,
    Protocol,
    Sequence,
    TypeVar,
    Union,
)

from asyncpg import (  # type: ignore
    Connection,
    Pool,
)
from pypika import (  # type: ignore
    Criterion,
    CustomFunction,
    Not,
    Order,
    Parameter,
    PostgreSQLQuery,
    Table,
    functions as fn,
)
from pypika.enums import Comparator  # type: ignore
from pypika.terms import (  # type: ignore
    BasicCriterion,
    Term,
)

from .core import F
from .core import FilterType
from .core import NotFound


T = TypeVar("T")


class _PostgreSQLMatching(Comparator):
    regexp = " ~ "
    not_regexp = " !~ "
    iregexp = " ~* "
    not_iregexp = " !~* "
    inet_in = " << "
    overlap = " && "


class _Regexp(BasicCriterion):
    def __init__(self, term: Term, expr: str) -> None:
        super().__init__(_PostgreSQLMatching.regexp, term, term.wrap_constant(expr))


class _NotRegexp(BasicCriterion):
    def __init__(self, term: Term, expr: str) -> None:
        super().__init__(_PostgreSQLMatching.not_regexp, term, term.wrap_constant(expr))


class _IRegexp(BasicCriterion):
    def __init__(self, term: Term, expr: str) -> None:
        super().__init__(_PostgreSQLMatching.iregexp, term, term.wrap_constant(expr))


class _NotIRegexp(BasicCriterion):
    def __init__(self, term: Term, expr: str) -> None:
        super().__init__(
            _PostgreSQLMatching.not_iregexp, term, term.wrap_constant(expr)
        )


class _Overlap(BasicCriterion):
    def __init__(self, term: Term, expr: str) -> None:
        super().__init__(_PostgreSQLMatching.overlap, term, term.wrap_constant(expr))


class _InetIn(BasicCriterion):
    inet = CustomFunction("inet", ["value"])

    def __init__(self, term: Term, expr: str) -> None:
        super().__init__(
            _PostgreSQLMatching.inet_in,
            self.inet(term),
            term.wrap_constant(self.inet(expr)),
        )


class PostgresRepo(Generic[T]):
    """Implementation of the :class:`misery.Repo` protocol
    that uses asyncpg to communicate with PostgreSQL.
    """

    id_field = "id"
    """Name of the ID field of an entity."""

    @property
    @abstractmethod
    def table(self) -> Table:
        """Main table of the repository.
        It is used to autogenerate SQL queries.
        """
        ...

    def __init__(self, conn: Union[Connection, Pool]) -> None:
        """:param conn: Connection or pool that
        will be used for interaction with the database.
        """
        self._conn = conn

    @property
    def conn(self) -> Union[Connection, Pool]:
        """Connection or pool that is used
        for interaction with the database.

        Use this property in your custom methods.
        """
        return _current_conn.get() or self._conn

    @property
    def query(self) -> PostgreSQLQuery:
        """Query to select records from the database.

        Override this property when you need
        something more complex than selecting
        all columns from the main table
        of the repository.
        """
        return PostgreSQLQuery.from_(self.table).select("*")

    @property
    def _entity_type(self) -> type:
        return typing.get_args(self.__orig_bases__[0])[0]  # type: ignore

    def load(self, record: dict) -> T:
        """Map a database record to an entity.

        Override this method if needed.
        """
        return self._entity_type(**record)

    def dump(self, entity: T) -> dict:
        """Map an entity to a database record.

        Override this method if needed.
        """
        return dataclasses.asdict(entity)

    async def fetch_one(self, query: PostgreSQLQuery) -> T:
        """Find a record in the database
        and map it to an entity.
        """
        data = await self.conn.fetchrow(str(query))

        if data is None:
            raise NotFound

        return self.load(dict(data))

    async def fetch_many(self, query: PostgreSQLQuery) -> Iterable[T]:
        """Find multiple records in the database
        and map them to entities.
        """
        records = await self.conn.fetch(str(query))
        return map(lambda x: self.load(dict(x)), records)

    async def add(self, entity: T) -> None:
        data = self.dump(entity)
        query = (
            PostgreSQLQuery.into(self.table)
            .columns(*data.keys())
            .insert(*data.values())
        )
        await self.conn.execute(str(query))
        await self.after_add(entity)

    async def after_add(self, entity: T) -> None:
        """Action after adding a new entity.

        By default, this method doesn't do anything.
        Override it to your liking.
        """
        pass

    async def add_many(self, entities: Iterable[T]) -> None:
        ientities = iter(entities)
        first = next(ientities)
        data = self.dump(first)

        query = (
            PostgreSQLQuery.into(self.table)
            .columns(*data.keys())
            .insert(*[Parameter(f"${n}") for n in range(1, len(data) + 1)])
        )

        await self.conn.executemany(
            str(query),
            itertools.chain(
                [tuple(data.values())],
                (tuple(self.dump(x).values()) for x in ientities),
            ),
        )

    async def get(self, **kwargs: Any) -> T:
        query = self.query.limit(1)

        for k, v in kwargs.items():
            query = query.where(self.table[k] == v)

        return await self.fetch_one(query)

    async def get_for_update(self, **kwargs: Any) -> T:
        query = (
            PostgreSQLQuery.from_(self.table)
            .select(self.table[self.id_field])
            .for_update()
        )

        for k, v in kwargs.items():
            query = query.where(self.table[k] == v)

        await self.conn.fetchrow(str(query))
        return await self.get(**kwargs)

    async def get_many(
        self,
        filters: Sequence[F] = (),
        order: Sequence[str] = (),
        limit: Optional[int] = None,
        page: int = 1,
        offset: int = 0,
    ) -> Iterable[T]:
        query = self.query

        if limit is not None:
            query = query.limit(limit).offset((page - 1) * limit + offset)
        elif offset > 0:
            query = query.offset(offset)

        for field in order:
            if field.startswith("-"):
                query = query.orderby(field[1:], order=Order.desc)
            else:
                query = query.orderby(field)

        for f in filters:
            criterion = self._filter_to_criterion(f)
            query = query.where(criterion)

        return await self.fetch_many(query)

    def _filter_to_criterion(self, f: F) -> Criterion:
        if f.type == FilterType.OR:
            return Criterion.any([self._filter_to_criterion(ff) for ff in f.value])
        elif f.type == FilterType.AND:
            return Criterion.all([self._filter_to_criterion(ff) for ff in f.value])

        column = self.table[f.field]
        criterion: Any = None

        if f.type == FilterType.EQ:
            if f.value is None:
                criterion = column.isnull()
            else:
                criterion = column == f.value
        elif f.type == FilterType.NEQ:
            if f.value is None:
                criterion = column.notnull()
            else:
                criterion = column != f.value
        elif f.type == FilterType.LT:
            criterion = column < f.value
        elif f.type == FilterType.GT:
            criterion = column > f.value
        elif f.type == FilterType.LTE:
            criterion = column <= f.value
        elif f.type == FilterType.GTE:
            criterion = column >= f.value
        elif f.type == FilterType.STARTSWITH:
            criterion = column.like(f"{f.value}%")
        elif f.type == FilterType.NSTARTSWITH:
            criterion = column.not_like(f"{f.value}%")
        elif f.type == FilterType.ENDSWITH:
            criterion = column.like(f"%{f.value}")
        elif f.type == FilterType.NENDSWITH:
            criterion = column.not_like(f"%{f.value}")
        elif f.type == FilterType.ISTARTSWITH:
            criterion = column.ilike(f"{f.value}%")
        elif f.type == FilterType.NISTARTSWITH:
            criterion = column.not_ilike(f"{f.value}%")
        elif f.type == FilterType.IENDSWITH:
            criterion = column.ilike(f"%{f.value}")
        elif f.type == FilterType.NIENDSWITH:
            criterion = column.not_ilike(f"%{f.value}")
        elif f.type == FilterType.CONTAINS:
            criterion = column.like(f"%{f.value}%")
        elif f.type == FilterType.NCONTAINS:
            criterion = column.not_like(f"%{f.value}%")
        elif f.type == FilterType.ICONTAINS:
            criterion = column.ilike(f"%{f.value}%")
        elif f.type == FilterType.NICONTAINS:
            criterion = column.not_ilike(f"%{f.value}%")
        elif f.type == FilterType.IN:
            criterion = column.isin(f.value)
        elif f.type == FilterType.NIN:
            criterion = column.notin(f.value)
        elif f.type == FilterType.MATCHES:
            criterion = _Regexp(column, f.value)
        elif f.type == FilterType.NMATCHES:
            criterion = _NotRegexp(column, f.value)
        elif f.type == FilterType.IMATCHES:
            criterion = _IRegexp(column, f.value)
        elif f.type == FilterType.NIMATCHES:
            criterion = _NotIRegexp(column, f.value)
        elif f.type == FilterType.IPIN:
            criterion = _InetIn(column, f.value)
        elif f.type == FilterType.NIPIN:
            criterion = Not(_InetIn(column, f.value))
        elif f.type == FilterType.HASANY:
            criterion = _Overlap(column, f.value)

        if f.not_:
            return Not(criterion)

        return criterion

    async def get_first(
        self, filters: Sequence[F] = (), order: Sequence[str] = ()
    ) -> T:
        try:
            return list(await self.get_many(filters, order=order, limit=1))[0]
        except IndexError:
            raise NotFound

    async def update(self, entity: T) -> None:
        query = PostgreSQLQuery.update(self.table)
        record = self.dump(entity)

        for k, v in record.items():
            query = query.set(k, v)

        query = query.where(self.table[self.id_field] == getattr(entity, self.id_field))
        result = await self.conn.execute(str(query))

        if result == "UPDATE 0":
            raise NotFound

        await self.after_update(entity)

    async def after_update(self, entity: T) -> None:
        """Action after update.

        By default, this method doesn't do anything.
        Override it to your liking.
        """
        pass

    async def delete(self, **kwargs: Any) -> None:
        query = PostgreSQLQuery.from_(self.table).delete()

        for k, v in kwargs.items():
            query = query.where(self.table[k] == v)

        await self.conn.execute(str(query))

    async def exists(self, **kwargs: Any) -> bool:
        query = PostgreSQLQuery.from_(self.table).select(1).limit(1)

        for k, v in kwargs.items():
            query = query.where(self.table[k] == v)

        data = await self.conn.fetchval(str(query))
        return data is not None

    async def count(self, **kwargs: Any) -> int:
        query = PostgreSQLQuery.from_(self.table).select(fn.Count("*"))

        for k, v in kwargs.items():
            query = query.where(self.table[k] == v)

        return await self.conn.fetchval(str(query))

    async def count_filtered(self, filter_: F) -> int:
        criterion = self._filter_to_criterion(filter_)
        query = self.query.where(criterion).distinct()
        query = PostgreSQLQuery.from_(query).select(fn.Count(1))
        return await self.conn.fetchval(str(query))


_current_transaction = ContextVar("_current_transaction", default=None)
_current_conn = ContextVar("_current_conn", default=None)


class PostgresTransactionManager:
    """Implementation of the :class:`misery.TransactionManager` protocol
    that uses asyncpg to communicate with PostgreSQL.
    """

    def __init__(self, conn: Union[Connection, Pool]) -> None:
        if isinstance(conn, Pool):
            self._pool, self._conn = conn, None
        else:
            self._pool, self._conn = None, conn

    @property
    def conn(self) -> Connection:
        if self._pool is None:
            return self._conn

        return _current_conn.get()

    async def __aenter__(self) -> None:
        if self._pool is not None:
            _current_conn.set(await self._pool.acquire())

        t = self.conn.transaction()
        await t.__aenter__()
        _current_transaction.set(t)

    async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        t = _current_transaction.get()

        if t is not None:
            await t.__aexit__(exc_type, exc, tb)
            _current_transaction.set(None)

        if _current_conn.get() is not None:
            await self._pool.release(_current_conn.get())
            _current_conn.set(None)



from __future__ import annotations

import enum
from contextlib import AbstractAsyncContextManager
from enum import Enum
from functools import partial
from typing import Any
from typing import Iterable
from typing import Optional
from typing import Protocol
from typing import Sequence
from typing import TypeVar
from typing import Union
from typing import ForwardRef


class FilterType(Enum):
    EQ = enum.auto()
    NEQ = enum.auto()
    LT = enum.auto()
    LTE = enum.auto()
    GT = enum.auto()
    GTE = enum.auto()
    CONTAINS = enum.auto()
    ICONTAINS = enum.auto()
    NCONTAINS = enum.auto()
    NICONTAINS = enum.auto()
    STARTSWITH = enum.auto()
    ISTARTSWITH = enum.auto()
    NSTARTSWITH = enum.auto()
    NISTARTSWITH = enum.auto()
    ENDSWITH = enum.auto()
    NENDSWITH = enum.auto()
    IENDSWITH = enum.auto()
    NIENDSWITH = enum.auto()
    MATCHES = enum.auto()
    IMATCHES = enum.auto()
    NMATCHES = enum.auto()
    NIMATCHES = enum.auto()
    IN = enum.auto()
    NIN = enum.auto()
    IPIN = enum.auto()
    NIPIN = enum.auto()
    HASANY = enum.auto()
    OR = enum.auto()
    AND = enum.auto()


class F:
    """Object used for filtering entities
    before getting them from a repository."""

    def __init__(
        self, type: FilterType, field: str, value: Any, not_: bool = False
    ) -> None:
        self.type = type
        self.field = field
        self.value = value
        self.not_ = not_

    def __repr__(self) -> str:
        return f"F(type={self.type.__repr__()}, field={self.field.__repr__()}, value={self.value.__repr__()})"

    def __invert__(self) -> F:
        return F(
            type=self.type,
            field=self.field,
            value=self.value,
            not_=True,
        )

    @classmethod
    def eq(cls, field: str, value: Any) -> F:
        """Create a filter to find entities
        with a specified field value.
        """
        return cls(FilterType.EQ, field, value)

    @classmethod
    def lt(cls, field: str, value: Any) -> F:
        """Create a filter to find entities
        with a field value that is
        less than a given one.
        """
        return cls(FilterType.LT, field, value)

    @classmethod
    def lte(cls, field: str, value: Any) -> F:
        """Create a filter to find entities
        with a field value that is
        less than or equal to a given one.
        """
        return cls(FilterType.LTE, field, value)

    @classmethod
    def gt(cls, field: str, value: Any) -> F:
        """Create a filter to find entities
        with a field value that is
        greater than a given one.
        """
        return cls(FilterType.GT, field, value)

    @classmethod
    def gte(cls, field: str, value: Any) -> F:
        """Create a filter to find entities
        with a field value that is
        greater than or equal to a given one.
        """
        return cls(FilterType.GTE, field, value)

    @classmethod
    def neq(cls, field: str, value: Any) -> F:
        """Create a filter to find entities
        with a field value that is
        not equal to a given one.
        """
        return cls(FilterType.NEQ, field, value)

    @classmethod
    def contains(cls, field: str, value: Any) -> F:
        """Create a filter to find entities
        whose field contains a given value.

        Case-sensitive.
        """
        return cls(FilterType.CONTAINS, field, value)

    @classmethod
    def icontains(cls, field: str, value: str) -> F:
        """Create a filter to find entities
        whose field contains a given value.

        Case-insensitive.
        """
        return cls(FilterType.ICONTAINS, field, value)

    @classmethod
    def ncontains(cls, field: str, value: Any) -> F:
        """Create a filter to find entities
        whose field doesn't contain a given value.

        Case-sensitive.
        """
        return cls(FilterType.NCONTAINS, field, value)

    @classmethod
    def nicontains(cls, field: str, value: str) -> F:
        """Create a filter to find entities
        whose field doesn't contain a given value.

        Case-insensitive.
        """
        return cls(FilterType.NICONTAINS, field, value)

    @classmethod
    def startswith(cls, field: str, value: str) -> F:
        """Create a filter to find entities
        whose field starts with a given value.

        Case-sensitive.
        """
        return cls(FilterType.STARTSWITH, field, value)

    @classmethod
    def istartswith(cls, field: str, value: str) -> F:
        """Create a filter to find entities
        whose field starts with a given value.

        Case-insensitive.
        """
        return cls(FilterType.ISTARTSWITH, field, value)

    @classmethod
    def nstartswith(cls, field: str, value: str) -> F:
        """Create a filter to find entities
        whose field doesn't start with a given value.

        Case-sensitive.
        """
        return cls(FilterType.NSTARTSWITH, field, value)

    @classmethod
    def nistartswith(cls, field: str, value: str) -> F:
        """Create a filter to find entities
        whose field doesn't start with a given value.

        Case-insensitive.
        """
        return cls(FilterType.NISTARTSWITH, field, value)

    @classmethod
    def endswith(cls, field: str, value: str) -> F:
        """Create a filter to find entities
        whose field ends with a given value.

        Case-sensitive.
        """
        return cls(FilterType.ENDSWITH, field, value)

    @classmethod
    def nendswith(cls, field: str, value: str) -> F:
        """Create a filter to find entities
        whose field doesn't end with a given value.

        Case-sensitive.
        """
        return cls(FilterType.NENDSWITH, field, value)

    @classmethod
    def iendswith(cls, field: str, value: str) -> F:
        """Create a filter to find entities
        whose field value ends with a given value.

        Case-insensitive.
        """
        return cls(FilterType.IENDSWITH, field, value)

    @classmethod
    def niendswith(cls, field: str, value: str) -> F:
        """Create a filter to find entities
        whose field value doesn't end with a given value.

        Case-insensitive.
        """
        return cls(FilterType.NIENDSWITH, field, value)

    @classmethod
    def matches(cls, field: str, value: str) -> F:
        """Create a filter to find entities
        whose field matches
        a regular expression in the ``value`` parameter.

        Case-sensitive.
        """
        return cls(FilterType.MATCHES, field, value)

    @classmethod
    def imatches(cls, field: str, value: str) -> F:
        """Create a filter to find entities
        whose field value matches
        a regular expression in the ``value`` parameter.

        Case-insensitive.
        """
        return cls(FilterType.IMATCHES, field, value)

    @classmethod
    def nmatches(cls, field: str, value: str) -> F:
        """Create a filter to find entities
        whose field value doesn't match
        a regular expression in the ``value`` parameter.

        Case-sensitive.
        """
        return cls(FilterType.NMATCHES, field, value)

    @classmethod
    def nimatches(cls, field: str, value: str) -> F:
        """Create a filter to find entities
        whose field value doesn't match
        a regular expression in the ``value`` parameter.

        Case-insensitive.
        """
        return cls(FilterType.NIMATCHES, field, value)

    @classmethod
    def in_(cls, field: str, value: Any) -> F:
        """Create a filter to find entities
        whose field is within a given value.

        Case-sensitive.
        """
        return cls(FilterType.IN, field, value)

    @classmethod
    def nin(cls, field: str, value: Any) -> F:
        """Create a filter to find entities
        whose field is not within a given value.

        Case-sensitive.
        """
        return cls(FilterType.NIN, field, value)

    @classmethod
    def ipin(cls, field: str, subnet: str) -> F:
        """Create a filter to find entities
        whose field is an IP address contained within
        a given subnet.
        """
        return cls(FilterType.IPIN, field, subnet)

    @classmethod
    def nipin(cls, field: str, subnet: str) -> F:
        """Create a filter to find entities
        whose field is an IP address not contained within
        a given subnet.
        """
        return cls(FilterType.NIPIN, field, subnet)

    @classmethod
    def hasany(cls, field: str, value: Sequence) -> F:
        """Create a filter to find entities
        whose array-like field contains any
        value from a given sequence.
        """
        return cls(FilterType.HASANY, field, value)

    @classmethod
    def or_(cls, *filters: F) -> F:
        """Create a filter that combines multiple
        filters with the "OR" operator.
        """
        return cls(FilterType.OR, "", filters)

    @classmethod
    def and_(cls, *filters: F) -> F:
        """Create a filter that combines multiple
        filters with the "AND" operator.
        """
        return cls(FilterType.AND, "", filters)

    def __or__(self, other: F) -> F:
        return F.or_(self, other)

    def __and__(self, other: F) -> F:
        return F.and_(self, other)


T = TypeVar("T")


class Repo(Protocol[T]):
    """Protocol for all repositories to implement."""

    async def add(self, entity: T) -> None:
        """Add a new entity to the repository.

        :param entity: An entity to add.

        """
        ...

    async def add_many(self, entities: Iterable[T]) -> None:
        """Add multiple entities to the repository.

        :param entities: Entities to add.
        """
        ...

    async def get(self, **kwargs: Any) -> T:
        """Get an entity.
        Raise :exc:`misery.NotFound` if the entity is missing.

        :param kwargs: Lookup parameters as field-value pairs.
        """
        ...

    async def get_for_update(self, **kwargs: Any) -> T:
        """Get an entity and lock it for update.
        Raise :exc:`misery.NotFound` if the entity is missing.

        :param kwargs: Lookup parameters as field-value pairs.
        """
        ...

    async def get_many(
        self,
        filters: Sequence[F] = (),
        order: Sequence[str] = (),
        limit: Optional[int] = None,
        page: int = 1,
        offset: int = 0,
    ) -> Iterable[T]:
        """Get multiple entities.

        :param filters: Sequence of filters.
        :param order: Sequence of fields by which to sort entities.
            If a field starts with the "-" character,
            entities for the field will be shown in descending order.
        :param limit: Maximum number of entities to show on the page.
        :param page: Page number. Only when the ``limit`` parameter
            is used.
        :param offset: Number of entities to skip.
        """
        ...

    async def get_first(
        self, filters: Sequence[F] = (), order: Sequence[str] = ()
    ) -> T:
        """Get the first entity matching given filters.
        Raise :exc:`misery.NotFound` if the entity is missing.

        :param filters: Sequence of filters.
        :param order: Sequence of fields by which to sort entities.
            If a field starts with the "-" character,
            entities for the field will be shown in descending order.
        """
        ...

    async def update(self, entity: T) -> None:
        """Save an updated entity.
        Raise :exc:`misery.NotFound` if the entity
        is not present in the repository.

        :param entity: An updated entity.
        """
        ...

    async def delete(self, **kwargs: Any) -> None:
        """Delete entities. When ``kwargs`` is empty, delete
        all.

        :param kwargs: Lookup parameters as field-value pairs.
        """
        ...

    async def exists(self, **kwargs: Any) -> bool:
        """Check if there is an entity for given lookup parameters.

        :param kwargs: Lookup parameters as field-value pairs.
        """
        ...

    async def count(self, **kwargs: Any) -> int:
        """Count entities that match given lookup parameters.
        If there are no lookup parameters, count all entities in
        the repository.

        :param kwargs: Lookup parameters as field-value pairs.
        """
        ...

    async def count_filtered(self, filter_: F) -> int:
        """Count entities that match a given filter.
        """
        ...


class NotFound(Exception):
    """Error to raise when an entity cannot be found
    in the repository."""

    pass


class QueryError(Exception):
    """Error to raise when a query to the database
    cannot be processed."""

    def __init__(self, message: str, query: str) -> None:
        """:param message: Message from the database system.
        :param query: Query that cannot be processed.
        """
        self.message = message
        self.query = query

    def __str__(self) -> str:
        return f"QueryError(message='{self.message}', query='{self.query}')"


class TransactionManager(Protocol):
    """Protocol of an object that is used
    for transaction management."""

    async def __aenter__(self) -> None:
        """Start a new transaction."""
        ...

    async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        """Commit or rollback a transaction."""
        ...



from __future__ import annotations

import copy
import operator
import re
from abc import abstractmethod
from functools import reduce
from functools import total_ordering
from typing import Any
from typing import Callable
from typing import Generic
from typing import Iterable
from typing import Optional
from typing import Sequence
from typing import TypeVar

import ipaddress

from .core import F, FilterType
from .core import NotFound


T = TypeVar("T")


@total_ordering
class _Desc:
    def __init__(self, value: Any):
        self.value = value

    def __lt__(self, other: Any) -> bool:
        return self.value > other.value

    def __eq__(self, other: Any) -> bool:
        return self.value == other.value


class DictRepo(Generic[T]):
    """Implementation of the :class:`misery.Repo` protocol.
    It uses a dictionary to store entities.
    """

    id_field = "id"
    """Name of the ID field of an entity."""

    @property
    @abstractmethod
    def key(self) -> str:
        """Key inside the ``storage`` dictionary.
        The value of this key is also a dictionary.
        The repository uses this nested dictionary
        to save entities.
        """
        ...

    def __init__(self, storage: dict) -> None:
        """:param storage: Dictionary to use as a storage space.
        A storage space can be shared between multiple repositories.
        """
        self.storage = storage
        self.storage[self.key] = self.storage.get(self.key, {})

    @property
    def data(self) -> dict[Any, T]:
        """Nested dictionary inside the ``storage`` dictionary.
        The repository uses this nested dictionary
        to save entities.
        """
        return self.storage[self.key]

    async def add(self, entity: T) -> None:
        self.data[getattr(entity, self.id_field)] = copy.deepcopy(entity)

    async def add_many(self, entities: Iterable[T]) -> None:
        for entity in entities:
            await self.add(entity)

    async def get(self, **kwargs: Any) -> T:
        for entity in self.data.values():
            if self._matches(entity, **kwargs):
                return copy.deepcopy(entity)

        raise NotFound

    async def get_for_update(self, **kwargs: Any) -> T:
        return await self.get(**kwargs)

    def _matches(self, entity: T, **kwargs: Any) -> bool:
        return all(getattr(entity, k) == v for k, v in kwargs.items())

    async def get_many(
        self,
        filters: Sequence[F] = (),
        order: Sequence[str] = (),
        limit: Optional[int] = None,
        page: int = 1,
        offset: int = 0,
    ) -> Iterable[T]:
        result = self._apply_filters(self.data.values(), filters)

        def sorting_key(x: Any) -> tuple:
            return tuple(
                _Desc(self._get_field(x, f[1:]))
                if f.startswith("-")
                else self._get_field(x, f)
                for f in order
            )

        result = sorted(result, key=sorting_key)

        if limit is not None:
            result = result[(page - 1) * limit + offset : page * limit + offset]

        return map(copy.deepcopy, result)

    _FILTERS_MAP = {
        FilterType.EQ: operator.eq,
        FilterType.NEQ: operator.ne,
        FilterType.LT: operator.lt,
        FilterType.LTE: operator.le,
        FilterType.GT: operator.gt,
        FilterType.GTE: operator.ge,
        FilterType.CONTAINS: lambda s, v: v in s,
        FilterType.NCONTAINS: lambda s, v: v not in s,
        FilterType.ICONTAINS: lambda s, v: v.lower() in s.lower(),
        FilterType.NICONTAINS: lambda s, v: v.lower() not in s.lower(),
        FilterType.STARTSWITH: lambda s, v: s.startswith(v),
        FilterType.NSTARTSWITH: lambda s, v: not s.startswith(v),
        FilterType.ISTARTSWITH: lambda s, v: s.lower().startswith(v.lower()),
        FilterType.NISTARTSWITH: lambda s, v: not s.lower().startswith(v.lower()),
        FilterType.ENDSWITH: lambda s, v: s.endswith(v),
        FilterType.NENDSWITH: lambda s, v: not s.endswith(v),
        FilterType.IENDSWITH: lambda s, v: s.lower().endswith(v.lower()),
        FilterType.NIENDSWITH: lambda s, v: not s.lower().endswith(v.lower()),
        FilterType.IN: lambda s, v: s in v,
        FilterType.NIN: lambda s, v: s not in v,
        FilterType.MATCHES: lambda s, v: bool(re.match(v, s)),
        FilterType.NMATCHES: lambda s, v: not re.match(v, s),
        FilterType.IMATCHES: lambda s, v: bool(re.match(v, s, re.IGNORECASE)),
        FilterType.NIMATCHES: lambda s, v: not re.match(v, s, re.IGNORECASE),
        FilterType.IPIN: lambda s, v: ipaddress.ip_address(s)
        in ipaddress.ip_network(v),
        FilterType.NIPIN: lambda s, v: ipaddress.ip_address(s)
        not in ipaddress.ip_network(v),
        FilterType.HASANY: lambda s, v: any(i in s for i in v),
    }

    def _filter_to_op(self, f: F) -> Callable:
        op = self._FILTERS_MAP[f.type]
        if f.not_:
            return lambda s, v: not op(s, v)
        return op

    def _get_field(self, obj: Any, field: str) -> Any:
        return reduce(lambda x, y: getattr(x, y), field.split("."), obj)

    def _apply_filters(self, entities: Iterable[T], filters: Sequence[F]) -> list[T]:
        result = list(entities)

        for f in filters:
            result = [x for x in result if self._check(f, x)]

        return result

    def _check(self, filter_: F, entity: T) -> bool:
        if filter_.type == FilterType.OR:
            return any(self._check(ff, entity) for ff in filter_.value)
        elif filter_.type == FilterType.AND:
            return all(self._check(ff, entity) for ff in filter_.value)
        else:
            op = self._filter_to_op(filter_)
            return op(self._get_field(entity, filter_.field), filter_.value)

    async def get_first(
        self, filters: Sequence[F] = (), order: Sequence[str] = ()
    ) -> T:
        try:
            return list(await self.get_many(filters, order=order, limit=1))[0]
        except IndexError:
            raise NotFound

    async def update(self, entity: T) -> None:
        id = getattr(entity, self.id_field)
        if id not in self.data:
            raise NotFound
        self.data[id] = copy.deepcopy(entity)

    async def delete(self, **kwargs: Any) -> None:
        for id, item in list(self.data.items()):
            if self._matches(item, **kwargs):
                del self.data[id]

    async def exists(self, **kwargs: Any) -> bool:
        return any(self._matches(entity, **kwargs) for entity in self.data.values())

    async def count(self, **kwargs: Any) -> int:
        if kwargs:
            return sum(
                1 for entity in self.data.values() if self._matches(entity, **kwargs)
            )
        return len(self.data)

    async def count_filtered(self, filter_: F) -> int:
        return sum(1 for _ in await self.get_many([filter_]))


class DictTransactionManager:
    """Implementation of the :class:`misery.TransactionManager` protocol.
    It uses a dictionary to store entities.
    """

    def __init__(self, storage: dict) -> None:
        self._storage = storage
        self._copy: dict = {}

    async def __aenter__(self) -> None:
        self._copy = copy.deepcopy(self._storage)

    async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        if exc is not None:
            self._storage.clear()
            self._storage.update(self._copy)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['misery']

package_data = \
{'': ['*']}

install_requires = \
['ipaddress>=1.0.23,<2.0.0']

extras_require = \
{'clickhouse': ['PyPika>=0.48.9,<0.49.0', 'aiohttp>=3.8.1,<4.0.0'],
 'postgres': ['PyPika>=0.48.9,<0.49.0', 'asyncpg>=0.25.0,<0.26.0']}

setup_kwargs = {
    'name': 'misery',
    'version': '0.6.0',
    'description': 'asyncio-friendly database toolkit',
    'long_description': '<p align="center">\n    <img src="https://github.com/meowmeowcode/misery/blob/clickhouse/docs/source/_static/misery.png" width="200" alt="misery" />\n</p>\n\n\n# Misery\n\nAn **asyncio**-friendly database toolkit that works well with **MyPy**.\n\n## Supported database systems\n\nAt the moment, PostgreSQL and ClickHouse are supported.\n\n## Documentation\n\nThe latest documentation: https://misery.readthedocs.io\n\n## Usage example\n\n```python\nfrom dataclasses import dataclass\nfrom uuid import UUID, uuid4\n\nimport asyncpg\nfrom pypika import Table\nfrom misery.postgres import PostgresRepo\n\n\nconn = await asyncpg.connect("postgresql://postgres:password@localhost/postgres")\n\nawait conn.execute(\n    """\n        CREATE TABLE users (\n            id uuid PRIMARY KEY,\n            name text NOT NULL UNIQUE\n        );\n    """\n)\n\n\n@dataclass\nclass User:\n    id: UUID\n    name: str\n\n\nclass UsersRepo(PostgresRepo[User]):\n    table = Table("users")\n\n\nusers_repo = UsersRepo(conn)\n\nuser_id = uuid4()\nbob = User(id=user_id, name="Bob")\nawait users_repo.add(bob)\n\nuser = await users_repo.get(id=user_id)\nassert user == bob\n```',
    'author': 'Anton Evdokimov',
    'author_email': 'meowmeowcode@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
